#define G_LOG_DOMAIN "ap-skill"

#include "ap-skill.h"

#include <glib.h>
#include <glib-object.h>

typedef struct _APSkillPrivate
{
    gchar    *name;
    gint      score;
    gchar    *modifier;
    gint      bonus;
    gboolean  proficient;
} APSkillPrivate;

struct _APSkill
{
    GObject parent_instance;
};

/**
 * SECTION:ap-skill
 * @short_description: Manages a single skill of a character
 * @title: APSkill
 * @include: adventure-pack/ap-skill.h
 *
 * #APSkill is an object for creating and managing a character's skill.
 * Allows the setting of various skill attributes.
 *
 * #APSkill becomes dynamic when it is registers with #APManager. By setting a
 * modifier and being connected to #APManager the #APSkill will receive updates
 * from #APManager changing it's score depending on the ability modifier from
 * an #APAbility whos name matches the skill modifier.
 *
 * # Usage {#usage}
 * ## Basic
 * |[<!-- language="C" -->
 * APSKill *skill;
 *
 * skill = g_object_new (AP_TYPE_SKILL,
 *                       "name", "Acrobatics",
 *                       "modifier", "Dexterity",
 *                       NULL);
 * ]|
 *
 * ## Add properties
 * |[<!-- language="C" -->
 * APSKill *skill;
 *
 * skill = ap_skill_new ("Acrobatics");
 * ap_skill_add_modifier (ability, "Dexterity");
 * ]|
 */

G_DEFINE_TYPE_WITH_PRIVATE (APSkill, ap_skill, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_NAME,
    PROP_SCORE,
    PROP_BONUS,
    PROP_MODIFIER,
    PROP_PROFICIENT,
    NUM_PROPERTIES
};

static GParamSpec *obj_properties[NUM_PROPERTIES] = { NULL, };

enum
{
    SCORE_CHANGED,
    MODIFER_CHANGED,
    N_SIGNALS
};

static guint signals[N_SIGNALS] = { 0, };

static void
score_changed (APSkill *self,
               gint     score)
{
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    if (priv->score == score)
        return;

    priv->score = score;

    g_signal_emit (self, signals[SCORE_CHANGED], 0, priv->score);
}

static void
modifier_changed (APSkill *self,
                  gchar   *modifier)
{
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    if (priv->modifier == modifier)
        return;

    priv->modifier = modifier;

    g_signal_emit (self, signals[MODIFER_CHANGED], 0, priv->modifier);
}

static void
ap_skill_set_property (GObject      *object,
                       guint         property_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
    APSkill *self = AP_SKILL (object);
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            priv->name = g_value_dup_string (value);
            break;

        case PROP_SCORE:
            score_changed (self, g_value_get_int (value));
            break;

        case PROP_BONUS:
            priv->bonus = g_value_get_int (value);
            break;

        case PROP_MODIFIER:
            modifier_changed (self, g_value_dup_string (value));
            break;

        case PROP_PROFICIENT:
            priv->proficient = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_skill_get_property (GObject      *object,
                       guint         property_id,
                       GValue       *value,
                       GParamSpec   *pspec)
{
    APSkill *self = AP_SKILL (object);
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            g_value_set_string (value, priv->name);
            break;

        case PROP_SCORE:
            g_value_set_int (value, priv->score);
            break;

        case PROP_BONUS:
            g_value_set_int (value, priv->bonus);
            break;

        case PROP_MODIFIER:
            g_value_set_string (value, priv->modifier);

            break;

        case PROP_PROFICIENT:
            g_value_set_boolean (value, priv->proficient);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_skill_finalize (GObject *object)
{
    APSkill *self = AP_SKILL (object);
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    g_free (priv->name);
    g_free (priv->modifier);

    G_OBJECT_CLASS (ap_skill_parent_class)->finalize (object);
}

static void
ap_skill_init (APSkill *self)
{
    APSkillPrivate *priv = ap_skill_get_instance_private (self);

    priv->name = NULL;
    priv->score = 0;
    priv->modifier = NULL;
    priv->bonus = 0;
    priv->proficient = FALSE;
}

static void
ap_skill_class_init (APSkillClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = ap_skill_set_property;
    object_class->get_property = ap_skill_get_property;
    object_class->finalize = ap_skill_finalize;

	/**
	 * APSkill:name:
	 *
	 * Name of the skill.
	 */
    obj_properties[PROP_NAME] =
        g_param_spec_string ("name",
                             "name",
                             "Name of the skill.",
                             NULL,
                             G_PARAM_READWRITE);

	/**
	 * APSkill:score:
	 *
	 * Score of the skill.
	 */
    obj_properties[PROP_SCORE] =
        g_param_spec_int ("score",
                          "score",
                          "The skill score level.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APSkill:bonus:
	 *
	 * Skill bonus of the ability.
	 */
    obj_properties[PROP_BONUS] =
        g_param_spec_int ("bonus",
                          "bonus",
                          "The skill bonus.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APSkill:modifier:
	 *
	 * Modifier of the skill.
	 */
    obj_properties[PROP_MODIFIER] =
        g_param_spec_string ("modifier",
                             "modifier",
                             "The ability id that modifies the skill.",
                             NULL,
                             G_PARAM_READWRITE);

	/**
	 * APSkill:proficient:
	 *
	 * Skill proficiency of the ability.
	 */
    obj_properties[PROP_PROFICIENT] =
        g_param_spec_boolean ("proficient",
                              "proficient",
                              "Is the character proficient with this skill?",
                              FALSE,
                              G_PARAM_READWRITE);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       obj_properties);

    /**
     * APSkill::score-changed:
     * @skill: an #APSkill
     * @score: the new score of @skill
     *
     * The ::score-changed signal is emitted each time the @skill<!-- -->'s
     * score has changed.
     */
    signals[SCORE_CHANGED] =
        g_signal_new ("score-changed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_INT);

    /**
     * APSkill::modifier-changed:
     * @skill: an #APSkill
     * @name: the name of the ability that modifies @skill
     *
     * The ::modifier-changed signal is emitted each time the @skill<!-- -->'s
     * modifier has changed.
     */
    signals[MODIFER_CHANGED] =
        g_signal_new ("modifier-changed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_skill_new:
 *
 * Creates a new #APSkill.
 *
 * Returns: (transfer full): an #APSkill instance
 */
APSkill *
ap_skill_new (void)
{
    return g_object_new (AP_TYPE_SKILL, NULL);
}

/**
 * ap_skill_new_with_name:
 * @name: the skill name for new #APSkill
 *
 * Creates a new #APSkill with @name.
 *
 * Returns: (transfer full): an #APSkill instance
 */
APSkill *
ap_skill_new_with_name (gchar *name)
{
    return g_object_new (AP_TYPE_SKILL,
                         "name", name,
                         NULL);
}

/**
 * ap_skill_set_name:
 * @self: an #APSkill
 * @name: the skill name for @self
 *
 * Sets the skill name for @self.
 *
 * ## Examples
 * * `Acrobatics`
 * * `Animal Handling`
 */
void
ap_skill_set_name (APSkill *self,
                   gchar   *name)
{
    GValue new_name = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));
    g_assert (name != NULL);
    g_assert (g_strcmp0 (name, "") != 0);

    g_value_init (&new_name, G_TYPE_STRING);
    g_value_set_string (&new_name, name);

    g_object_set_property (G_OBJECT (self), "name", &new_name);

    g_value_unset (&new_name);
}

/**
 * ap_skill_get_name:
 * @self: an #APSkill
 *
 * Gets the name of @self.
 *
 * Returns: (transfer full) (nullable): the skill name of @self, free with
 *          g_free()
 */
gchar *
ap_skill_get_name (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "name", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_score:
 * @self: am #APSkill
 * @score: the skill score for @self
 *
 * Sets the skill score for @self.
 */
void
ap_skill_set_score (APSkill *self,
                    gint     score)
{
    GValue new_score = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_score, G_TYPE_INT);
    g_value_set_int (&new_score, score);

    g_object_set_property (G_OBJECT (self), "score", &new_score);

    g_value_unset (&new_score);
}

/**
 * ap_skill_get_score:
 * @self: an #APSkill
 *
 * Gets the skill score of @self.
 *
 * Returns: (nullable): the skill score of @self
 */
gint
ap_skill_get_score (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "score", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_bonus:
 * @self: an #APSkill
 * @bonus: the skill score bonus for @self
 *
 * Sets the skill score bonus for @self.
 */
void
ap_skill_set_bonus (APSkill *self,
                    gint     bonus)
{
    GValue new_bonus = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_bonus, G_TYPE_INT);
    g_value_set_int (&new_bonus, bonus);

    g_object_set_property (G_OBJECT (self), "bonus", &new_bonus);

    g_value_unset (&new_bonus);
}

/**
 * ap_skill_get_bonus:
 * @self: an #APSkill
 *
 * Gets the skill score bonus of @self.
 *
 * Returns: (nullable): the skill score bonus of @self
 */
gint
ap_skill_get_bonus (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "bonus", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_modifier:
 * @self: an #APSkill
 * @modifier: the ability name that acts as the modifier for @self
 *
 * Sets the ability name that acts as the modifier for @self.
 */
void
ap_skill_set_modifier (APSkill *self,
                       gchar   *modifier)
{
    GValue new_modifier = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_modifier, G_TYPE_STRING);
    g_value_set_string (&new_modifier, modifier);

    g_object_set_property (G_OBJECT (self), "modifier", &new_modifier);

    g_value_unset (&new_modifier);
}

/**
 * ap_skill_get_modifier:
 * @self: An instance of #APSkill
 *
 * Gets the ability name that acts as the modifier of @self.
 *
 * Returns: (nullable): the name of an #APAbility that acts as the modifier
 *          for @self, free with g_free()
 */
gchar *
ap_skill_get_modifier (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "modifier", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_skill_set_proficient:
 * @self: an #APSkill
 * @proficient: %TRUE if character is proficient with @self
 *
 * Sets the skill score bonus for @self.
 */
void
ap_skill_set_proficient (APSkill  *self,
                         gboolean  proficient)
{
    GValue new_proficient = G_VALUE_INIT;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&new_proficient, G_TYPE_BOOLEAN);
    g_value_set_boolean (&new_proficient, proficient);

    g_object_set_property (G_OBJECT (self), "proficient", &new_proficient);

    g_value_unset (&new_proficient);
}

/**
 * ap_skill_get_proficient:
 * @self: an #APSkill
 *
 * Returns if the character is proficient with @self.
 *
 * Returns: (nullable): %TRUE if character is proficient with @self
 */
gboolean
ap_skill_get_proficient (APSkill *self)
{
    GValue val = G_VALUE_INIT;
    gboolean retval;

    g_assert (AP_IS_SKILL (self));

    g_value_init (&val, G_TYPE_BOOLEAN);

    g_object_get_property (G_OBJECT (self), "proficient", &val);

    retval = g_value_get_boolean (&val);

    g_value_unset (&val);

    return retval;
}
